extern crate sdl2;

use std::thread::sleep;
use std::time::Duration;

use bitvec::prelude::*;
use sdl2::event::Event;
use sdl2::keyboard::Keycode;
use sdl2::pixels::{Color, PixelFormatEnum};
use sdl2::rect::Rect;
use sdl2::render::{WindowCanvas, Texture};

pub mod cp437;
use cp437::GLYPHS;

const SCREEN_HEIGHT: u32 = 350;
const SCREEN_WIDTH: u32 = 640;

const EGA_COLORS: [Color; 16] = [
    Color::RGB(0x00, 0x00, 0x00), // Black
    Color::RGB(0x00, 0x00, 0xaa), // Blue
    Color::RGB(0x00, 0xaa, 0x00), // Green
    Color::RGB(0x00, 0xaa, 0xaa), // Cyan
    Color::RGB(0xaa, 0x00, 0x00), // Red
    Color::RGB(0xaa, 0x00, 0xaa), // Magenta
    Color::RGB(0xaa, 0x55, 0x00), // Brown
    Color::RGB(0xaa, 0xaa, 0xaa), // White / Light Gray
    Color::RGB(0x55, 0x55, 0x55), // Dark Gray / Bright Black
    Color::RGB(0x55, 0x55, 0xff), // Bright Blue
    Color::RGB(0x55, 0xff, 0x55), // Bright Green
    Color::RGB(0x55, 0xff, 0xff), // Bright Cyan
    Color::RGB(0xff, 0x55, 0x55), // Bright Red
    Color::RGB(0xff, 0x55, 0xff), // Bright Magenta
    Color::RGB(0xff, 0xff, 0x55), // Bright Yellow
    Color::RGB(0xff, 0xff, 0xff), // Bright White
];

pub fn textattr_to_int(back: u8, fore: u8, codepoint: u8, blink: bool) -> usize {
    let num = (usize::from(blink) << 15) +
        (usize::from(back) << 12) +
        (usize::from(fore) << 8) +
         usize::from(codepoint);
    num
}

pub fn int_to_textattr(num: usize) -> (u8, u8, u8, bool) {
    let blink = (num >> 15) != 0;
    let back = (num >> 12) as u8;
    let fore = ((num >> 8) & 15) as u8;
    let codepoint = (num & 255) as u8;
    (back, fore, codepoint, blink)
}

pub fn render(canvas: &mut WindowCanvas, color: Color, texture: &Texture) {
    canvas.set_draw_color(color);
    canvas.clear();

    match canvas.copy(texture, Rect::new(0, 0, 8, 14), Rect::new(0, 0, 8, 14)) {
        Ok(()) => (),
        Err(err) => panic!("Unable to copy texture to canvas! SDL Error: {}", err),
    };

    match canvas.copy(texture, Rect::new(0, 0, 8, 14), Rect::new(8, 0, 8, 14)) {
        Ok(()) => (),
        Err(err) => panic!("Unable to copy texture to canvas! SDL Error: {}", err),
    };

    match canvas.copy(texture, Rect::new(0, 0, 8, 14), Rect::new(16, 0, 8, 14)) {
        Ok(()) => (),
        Err(err) => panic!("Unable to copy texture to canvas! SDL Error: {}", err),
    };

    canvas.present();
}

pub fn main() {
    let sdl_context = match sdl2::init() {
        Ok(sdl_context) => sdl_context,
        Err(err) => panic!("SDL could not initialize! SDL_Error: {}", err),
    };

    let video_subsystem = match sdl_context.video() {
        Ok(video) => video,
        Err(err) => panic!("Could not obtain handle to the video subsystem! SDL_Error: {}", err),
    };

    let window = match video_subsystem.window("DOS Text Mode", SCREEN_WIDTH, SCREEN_HEIGHT)
        .position_centered()
        .opengl()
        .build() {
            Ok(window) => window,
            Err(err) => panic!("SDL could not create a window! SDL_Error: {}", err),
        };

    let mut canvas = match window.into_canvas().build() {
        Ok(canvas) => canvas,
        Err(err) => panic!("SDL could not create a canvas! SDL_Error: {}", err),
    };
    
    let texture_creator = canvas.texture_creator();
    let mut glyph_textures: Vec<Texture> = Vec::new();
    
    for tex_idx in 0..=32767 {
        let (back, fore, cp, _) = int_to_textattr(tex_idx);
        let mut texture = match texture_creator.create_texture_streaming(PixelFormatEnum::RGB24, 8, 14) {
            Ok(texture) => texture,
            Err(err) => panic!("SDL could not create texture. SDL Error: {}", err),
        };
        match texture.with_lock(None, |buffer: &mut [u8], pitch: usize| {
            for (y, line) in GLYPHS[cp as usize].iter().enumerate() {
                let bits = line.view_bits::<Msb0>();
                for (x, pixel) in bits.iter().enumerate() {
                    let offset = y * pitch + x * 3;
                    if pixel == false {
                        buffer[offset] = EGA_COLORS[back as usize].r;
                        buffer[offset + 1] = EGA_COLORS[back as usize].g;
                        buffer[offset + 2] = EGA_COLORS[back as usize].b;
                    } else {
                        buffer[offset] = EGA_COLORS[fore as usize].r;
                        buffer[offset + 1] = EGA_COLORS[fore as usize].g;
                        buffer[offset + 2] = EGA_COLORS[fore as usize].b;
                    }
                }
            }
        }) {
            Ok(()) => (),
            Err(err) => panic!("Unable to modify texture. SDL Error: {}", err),
        };
        glyph_textures.push(texture);
    }
    
    render(&mut canvas, Color::RGB(0, 0, 0), &glyph_textures[3762]);

    let mut event_pump = match sdl_context.event_pump() {
        Ok(event_pump) => event_pump,
        Err(err) => panic!("Could not obtain SDL2 Event Pump! SDL_Error: {}", err),
    };
    'running: loop {
        for event in event_pump.poll_iter() {
            match event {
                Event::Quit {..} |
                Event::KeyDown { keycode: Some(Keycode::Escape), .. } => {
                    break 'running
                },
                _ => {}
            }
        }

        render(&mut canvas, Color::RGB(0, 0, 0), &glyph_textures[3762]);
        sleep(Duration::new(0, 1_000_000_000u32 / 60));
    }
}
